@@ -1,12 +1,14 @@
-import mpython
-import music
-import socket
-import network
-import json
-import gc
+__version__ = "TinyWebIO v0.0.7"
+__author__ = "roadlabs@gmail.com"
+__license__ = "http://unlicense.org"
+
+import mpython, music, socket, network, json, gc
+from ubinascii import hexlify
+from time import time
+from urequests import request
+from machine import Timer, unique_id
 
 
-# Request Class
 class Request:   
     def __init__(self, socket=None):
         self.client_socket = socket
@@ -35,7 +37,9 @@ class Request:
     def parse(self):
         gc.collect()
         try:
-            req_data = self.client_socket.recv(4096).decode() #.strip()
+            req_data = self.client_socket.recv(4096).decode().replace('+', ' ')   # .strip()
+            if not req_data:
+                raise Exception('no data')
             req_datas = req_data.split('\r\n')
             firstline = self._unquote(req_datas[0])
             lastline = self._unquote(req_datas[-1])
@@ -54,11 +58,10 @@ class Request:
                 for p in params:
                     k, v = p.split('=')
                     self.form[k] = v.strip('"')
-        except Exception:
-            print("request data parsed failure!")
+        except Exception as e:
+            print("request data parsed failure!:%s" % e)
 
 
-# Response Class
 class Response:   
     def __init__(self, socket=None):
         self.client_socket = socket
@@ -80,22 +83,11 @@ class Response:
         gc.collect()
         firstline = self.response_state
         header = b'Content-Type: text/html; charset=utf-8\r\n'
-        template = b'''
-        <html>
-          <head>
-            <title>{title}</title>
-            <meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1.0, minimum-scale=1.0"
-            <meta charset='utf-8'>
-            <link rel="icon" href="data:;base64,=">
-            <style>{style}</style>
-          </head>
-          <body class='c'>
-            {content}
-          </body>
-        </html>
+        template = b'''<html><head><title>{title}</title><meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1.0, minimum-scale=1.0"><meta charset='utf-8'><link rel="icon" href="data:;base64,=">
+        <style>{style}</style></head><body class='c'><p><a href='/'>{home}</a></p>{content}</body></html>
         '''
-        style = "*+*{box-sizing:border-box;margin:.5em 0}@media(min-width:35em){.col{display:table-cell}.\31{width:5%}.\33{width:22%}.\34{width:30%}.\35{width:40%}.\32{width:15%}.row{display:table;border-spacing:1em 0}}.row,.w-100{width:100%}.card:focus,hr{outline:0;border:solid #fa0}.card,pre{padding:1em;border:solid #eee}.btn:hover,a:hover{opacity:.6}.c{max-width:60em;padding:1em;margin:auto;font:1em/1.6 nunito}a{color:#fa0;text-decoration:none}.btn.primary{color:#fff;background:#fa0;border:solid #fa0}pre{overflow:auto}td,th{padding:1em;text-align:left;border-bottom:solid #eee}.btn{cursor:pointer;padding:1em;letter-spacing:.1em;text-transform:uppercase;background:#fff;border:solid;font:.7em nunito}"
-        data = template.format(title=title, style=style, content=content)
+        style = "*+*{box-sizing:border-box;margin:.5em 0}@media(min-width:35em){.col{display:table-cell}.row{display:table;border-spacing:1em 0}}.row,.w-100{width:100%}.card:focus,hr{outline:0;border:solid #fa0}.card,pre{padding:1em;border:solid #eee}.btn:hover,a:hover{opacity:.6}.c{max-width:60em;padding:1em;margin:auto;font:1em/1.6 nunito}a{color:#fa0;text-decoration:none}.btn.primary{color:#fff;background:#fa0;border:solid #fa0}pre{overflow:auto}td,th{padding:1em;text-align:left;border-bottom:solid #eee}.btn{cursor:pointer;padding:1em;letter-spacing:.1em;text-transform:uppercase;background:#fff;border:solid;font:.7em nunito}"
+        data = template.format(title=title, style=style, home=__version__, content=content)
         body = b'\r\n%s' % data
         self.response_data = firstline + header + body
 
@@ -103,14 +95,15 @@ class Response:
         if self.response_data:
             try:
                 self.client_socket.write(self.response_data)
-            except Exception:
-                print("response data sent failure!")
+            except Exception as e:
+                print("response data sent failure!:%s" % e)
  
 
-# mPython Board Class
 class Board:
     def __init__(self):
-        pass
+        self.scope = {}
+        self.error = None
+        self.id = hexlify(unique_id()).decode()
 
     def _get_real_tag(self, raw_tag):
         real_tag = raw_tag
@@ -167,8 +160,12 @@ class Board:
                 elif pmode == 'analog':
                     pin = pfun(n, 4)
                     value = '%d' % pin.read_analog()
-        except Exception:
-            print("error for board data reading!")
+            elif tag.startswith('id'):
+                value = self.id
+            elif tag.startswith('time'):
+                value = time()
+        except Exception as e:
+            print("error for board data reading!:%s" % e)
         return ["VALUE", raw_tag, value]    
     
     def write(self, raw_tag, raw_value):
@@ -187,12 +184,14 @@ class Board:
                 oled = getattr(mpython, tag.strip())
                 values = value.split(':', 1)
                 method = values[0].strip()
-                vdata = values[1].strip()
-                if method.strip() == 'show':
-                    content, x, y = vdata.split(':')
+                vdata = values[-1].strip()
+                if method == 'show':
+                    content, x, y = vdata.split(',')
                     oled.DispChar(content, int(x.strip()), int(y.strip()))
                 elif method == 'fill':
                     oled.fill(int(vdata))
+                else:
+                    oled.DispChar(values[0].strip(), 0, 0)
                 oled.show()
             elif tag.startswith('buzz'):
                 bz = getattr(mpython, 'buzz')
@@ -203,6 +202,9 @@ class Board:
                     bz.on(freq) 
                 elif method.startswith('off'):
                     bz.off()
+                else:
+                    freq = method.split(':', 1)[0].strip()
+                    bz.on(int(freq))
             elif tag.startswith('music'):
                 method = value.strip()
                 if method.startswith('pitch'):
@@ -216,6 +218,11 @@ class Board:
                         if hasattr(music, tune_builtin):
                             tune = getattr(music, tune_builtin)
                     music.play(tune)
+            elif tag.startswith('servo'):
+                servo = getattr(mpython, 'Servo')
+                pin = int(tag[5:])
+                param = value.strip()
+                servo(pin).write_angle(int(param))
             elif tag.startswith('MPythonPin'):
                 pname, pmode, pnum = tag.split('_')
                 pfun = getattr(mpython, pname)
@@ -227,12 +234,14 @@ class Board:
                 elif pmode == 'analog':
                     pin = pfun(n, 3)
                     pin.write_analog(val)
-        except Exception:
-            print(" error for board data writing!")
+            elif tag.startswith('client'):
+                cmd = getattr(appclient, value.strip())
+                cmd()                
+        except Exception as e:
+            print(" error for board data writing!:%s" % e)
         return ["STORED", raw_tag, value]
 
 
-# Http Server Class
 class Server:
     def __init__(self):
         self.handlers = {}
@@ -257,8 +266,8 @@ class Server:
                 if iface.active():
                     suc_info = 'tinywebio started on http://'
                     print("%s%s:%d" % (suc_info, iface.ifconfig()[0], port))
-        except Exception:
-            print("appserver error occurred!")
+        except Exception as e:
+            print("appserver error occurred!:%s" % e)
         return self.server_socket
    
     def start(self, port=8888):
@@ -285,8 +294,8 @@ class Server:
 
         try:
             self.process_data(request, response, board)
-        except Exception:
-            print("client connection failure!")
+        except Exception as e:
+            print("client connection failure!:%s" % e)
         finally:
             self.client_socket.close()
             csock.close()
@@ -312,11 +321,98 @@ class Server:
             self.client_socket.close()
 
 
+class Remote:
+    def __init__(self, url, lasttask):
+        self.lasttask = lasttask
+        self.method = 'POST'
+        self.headers = {'Content-Type': 'application/x-www-form-urlencoded'}
+        self.puburl = '%s/storeavalue' % url
+        self.suburl = '%s/getvalue' % url
+
+    def pub(self, tag, value):
+        _val = value[-1]
+        _data = 'tag=%s&value=%s' % (tag, _val)
+        request(self.method, self.puburl, data=_data, headers=self.headers)
+
+    def sub(self, tag):
+        _data = 'tag=%s' % tag
+        _res = request(self.method, self.suburl, data=_data, headers=self.headers)
+        _val = json.loads(_res.text)[-1].strip('"')
+        value = None
+        if not _val == self.lasttask[tag]:
+            self.lasttask[tag] = _val
+            value = _val.split('_')[0]
+        return value
+
+class Client:
+    def __init__(self):
+        self.url = None
+        self.topic = {'pub': [], 'sub': []}
+        self.interval = 1000
+        self.tim = None
+        self.lasttask = {}
+        self.currtask = None
+        self.tasks = {}
+
+    def task(self, name):
+        def wrapper(func):
+            self.tasks[name] = func
+            return func
+        return wrapper
+
+    def setup(self, url=None, pub='', sub='', interval=1000):
+        self.url = 'http://%s' % url
+        self.topic['pub'] = pub.split(',')
+        self.topic['sub'] = sub.split(',')
+        for _topic in self.topic['sub']:
+            self.lasttask[_topic.strip()] = None
+        self.interval = interval
+        print('TinywebDB server address:%s/%s' % (self.url, hexlify(unique_id()).decode()))
+
+    def exec(self):
+        _board = Board()
+        _url = '%s/%s' % (self.url, _board.id)
+        _remote = Remote(_url, self.lasttask)
+        _topic = self.topic
+        _tasks = self.tasks
+        while True:
+            try:
+                for _id in _tasks.keys():
+                    _task = _tasks[_id]
+                    if _task:
+                        _task(_remote, _topic, _board)
+            except Exception as e:
+                print('executing task failure:%s' % e)
+
+            gc.collect()
+            yield
+
+    def start(self):
+        self.stop()
+        try:
+            # settime()
+            self.tim = Timer(1)
+            self.currtask = self.exec()
+            next(self.currtask)
+            if len(self.url) > 0 and (len(self.topic['pub']) > 0 or len(self.topic['sub']) > 0):
+                self.tim.init(period=self.interval, mode=Timer.PERIODIC, callback=lambda t: self.currtask.send(None))
+        except Exception as e:
+            print('starting client failure:%s' % e)
+
+    def stop(self):
+        if self.currtask:
+            self.currtask.close()
+        if self.tim:
+            self.tim.deinit()
+        gc.collect()
+
+appclient = Client()
 appserver = Server()
+gc.collect()
 
 @appserver.route('/')
 def index(request, response, board):
-    title = '\u638c\u63a7\u677f\u6d4b\u8bd5'
+    title = '%s' % __version__
     content = '''
     <p>\u529f\u80fd\u5217\u8868
       <ul>
@@ -378,5 +474,20 @@ def storeAValue(request, response, board):
         response.make_page(title, content)
         response.send()
 
+@appclient.task('publish')
+def publishTask(remote, topic, board):
+    for _tag in topic['pub']:
+        tag = _tag.strip()
+        val = board.read(tag)
+        remote.pub(tag, val)
+
+@appclient.task('subscribe')
+def subscribeTask(remote, topic, board):
+    for _tag in topic['sub']:
+        tag = _tag.strip()
+        val = remote.sub(tag)
+        if val:
+            board.write(tag, val)
+
 if __name__ == "__main__":
     appserver.start()
\ No newline at end of file
